@@ -1,10 +1,9 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2002-2006
- *      Sleepycat Software.  All rights reserved.
+ * Copyright (c) 2002,2008 Oracle.  All rights reserved.
  *
- * $Id: TruncateAndRemoveTest.java,v 1.12 2006/01/04 04:15:47 linda Exp $
+ * $Id: TruncateAndRemoveTest.java,v 1.29.2.1 2008/10/02 01:39:41 mark Exp $
  */
 
 package com.sleepycat.je.cleaner;
@@ -33,6 +32,7 @@ import com.sleepycat.je.Transaction;
 import com.sleepycat.je.config.EnvironmentParams;
 import com.sleepycat.je.dbi.DatabaseId;
 import com.sleepycat.je.dbi.DatabaseImpl;
+import com.sleepycat.je.dbi.EnvironmentImpl;
 import com.sleepycat.je.junit.JUnitThread;
 import com.sleepycat.je.log.DumpFileReader;
 import com.sleepycat.je.log.FileManager;
@@ -44,6 +44,9 @@ import com.sleepycat.je.util.TestUtils;
 import com.sleepycat.je.utilint.DbLsn;
 import com.sleepycat.je.utilint.TestHook;
 
+/**
+ * Test cleaning and utilization counting for database truncate and remove.
+ */
 public class TruncateAndRemoveTest extends TestCase {
 
     private static final String DB_NAME1 = "foo";
@@ -55,13 +58,18 @@ public class TruncateAndRemoveTest extends TestCase {
     static {
         FORCE_CHECKPOINT.setForce(true);
     }
-   
+
     private static final boolean DEBUG = false;
 
     private File envHome;
     private Environment env;
+    private EnvironmentImpl envImpl;
     private Database db;
+    private DatabaseImpl dbImpl;
     private JUnitThread junitThread;
+    private boolean fetchObsoleteSize;
+    private boolean truncateOrRemoveDone;
+    private boolean dbEviction;
 
     public TruncateAndRemoveTest() {
         envHome = new File(System.getProperty(TestUtils.DEST_DIR));
@@ -76,7 +84,7 @@ public class TruncateAndRemoveTest extends TestCase {
 
     public void tearDown()
         throws IOException, DatabaseException {
-  
+
         if (junitThread != null) {
             while (junitThread.isAlive()) {
                 junitThread.interrupt();
@@ -92,16 +100,20 @@ public class TruncateAndRemoveTest extends TestCase {
         } catch (Throwable e) {
             System.out.println("tearDown: " + e);
         }
-                
+
         try {
+            //*
             TestUtils.removeLogFiles("tearDown", envHome, true);
             TestUtils.removeFiles("tearDown", envHome, FileManager.DEL_SUFFIX);
+            //*/
         } catch (Throwable e) {
             System.out.println("tearDown: " + e);
         }
 
         db = null;
+        dbImpl = null;
         env = null;
+        envImpl = null;
         envHome = null;
     }
 
@@ -135,7 +147,19 @@ public class TruncateAndRemoveTest extends TestCase {
         DbInternal.disableParameterValidation(config);
         config.setConfigParam("je.log.fileMax", "4000");
 
+        /* Obsolete LN size counting is optional per test. */
+        if (fetchObsoleteSize) {
+            config.setConfigParam
+                (EnvironmentParams.CLEANER_FETCH_OBSOLETE_SIZE.getName(),
+                 "true");
+        }
+
         env = new Environment(envHome, config);
+        envImpl = DbInternal.envGetEnvironmentImpl(env);
+
+        config = env.getConfig();
+        dbEviction = config.getConfigParam
+            (EnvironmentParams.ENV_DB_EVICTION.getName()).equals("true");
     }
 
     /**
@@ -149,6 +173,20 @@ public class TruncateAndRemoveTest extends TestCase {
         dbConfig.setTransactional(envConfig.getTransactional());
         dbConfig.setAllowCreate(true);
         db = env.openDatabase(useTxn, dbName, dbConfig);
+        dbImpl = DbInternal.dbGetDatabaseImpl(db);
+    }
+
+    /**
+     * Closes the database.
+     */
+    private void closeDb()
+        throws DatabaseException {
+
+        if (db != null) {
+            db.close();
+            db = null;
+            dbImpl = null;
+        }
     }
 
     /**
@@ -157,13 +195,12 @@ public class TruncateAndRemoveTest extends TestCase {
     private void closeEnv()
         throws DatabaseException {
 
-        if (db != null) {
-            db.close();
-            db = null;
-        }
+        closeDb();
+
         if (env != null) {
             env.close();
             env = null;
+            envImpl = null;
         }
     }
 
@@ -176,18 +213,27 @@ public class TruncateAndRemoveTest extends TestCase {
         openEnv(true);
         openDb(null, DB_NAME1);
         writeAndCountRecords(null, RECORD_COUNT);
-        DatabaseId saveId = DbInternal.dbGetDatabaseImpl(db).getId();
-        db.close();
-        db = null;
+        DatabaseImpl saveDb = dbImpl;
+        DatabaseId saveId = dbImpl.getId();
+        closeDb();
 
         Transaction txn = env.beginTransaction(null, null);
         truncate(txn, true);
         ObsoleteCounts beforeCommit = getObsoleteCounts();
         txn.commit();
+        truncateOrRemoveDone = true;
 
-        verifyUtilization(beforeCommit, 
-                          RECORD_COUNT + 2, // LNs, + MapLN + previous NameLN
-                          15);              // 1 root, 2 INs, 12 BINs
+        /* Make sure use count is decremented when we commit. */
+        assertDbInUse(saveDb, false);
+        openDb(null, DB_NAME1);
+        saveDb = dbImpl;
+        closeDb();
+        assertDbInUse(saveDb, false);
+
+        verifyUtilization(beforeCommit,
+                          RECORD_COUNT + // LNs
+                          3,   // prev MapLN + deleted MapLN + prev NameLN
+                          15); // 1 root, 2 INs, 12 BINs
 
         closeEnv();
         batchCleanAndVerify(saveId);
@@ -202,22 +248,29 @@ public class TruncateAndRemoveTest extends TestCase {
         openEnv(true);
         openDb(null, DB_NAME1);
         writeAndCountRecords(null, RECORD_COUNT);
-        db.close();
-        db = null;
+        DatabaseImpl saveDb = dbImpl;
+        closeDb();
 
         Transaction txn = env.beginTransaction(null, null);
         truncate(txn, true);
         ObsoleteCounts beforeAbort = getObsoleteCounts();
         txn.abort();
 
-        /* 
+        /* Make sure use count is decremented when we abort. */
+        assertDbInUse(saveDb, false);
+        openDb(null, DB_NAME1);
+        saveDb = dbImpl;
+        closeDb();
+        assertDbInUse(saveDb, false);
+
+        /*
          * The obsolete count should include the records inserted after
          * the truncate.
          */
         verifyUtilization(beforeAbort,
                           /* 1 new nameLN, 2 copies of MapLN for new db */
-                           3, 
-                           0); 
+                           3,
+                           0);
 
         /* Reopen, db should be populated. */
         openDb(null, DB_NAME1);
@@ -234,8 +287,7 @@ public class TruncateAndRemoveTest extends TestCase {
         openEnv(true);
         openDb(null, DB_NAME1);
         writeAndCountRecords(null, RECORD_COUNT);
-        db.close();
-        db = null;
+        closeDb();
 
         Transaction txn = env.beginTransaction(null, null);
         truncate(txn, true);
@@ -243,21 +295,36 @@ public class TruncateAndRemoveTest extends TestCase {
         /* populate the database with some more records. */
         openDb(txn, DB_NAME1);
         writeAndCountRecords(txn, RECORD_COUNT/4);
-        DatabaseId saveId = DbInternal.dbGetDatabaseImpl(db).getId();
-        db.close();
-        db = null;
+        DatabaseImpl saveDb = dbImpl;
+        DatabaseId saveId = dbImpl.getId();
+        closeDb();
         ObsoleteCounts beforeAbort = getObsoleteCounts();
         txn.abort();
 
-        /* 
+        /*
+         * We set truncateOrRemoveDone to true (meaning that per-DB utilization
+         * will not be verified) even though the txn was aborted because the
+         * discarded new DatabaseImpl will not be counted yet includes INs and
+         * LNs from the operations above.
+         */
+        truncateOrRemoveDone = true;
+
+        /* Make sure use count is decremented when we abort. */
+        assertDbInUse(saveDb, false);
+        openDb(null, DB_NAME1);
+        saveDb = dbImpl;
+        closeDb();
+        assertDbInUse(saveDb, false);
+
+        /*
          * The obsolete count should include the records inserted after
          * the truncate.
          */
         verifyUtilization(beforeAbort,
-                          /* newly inserted LNs, 1 new nameLN, 
+                          /* newly inserted LNs, 1 new nameLN,
                            * 2 copies of MapLN for new db */
-                          (RECORD_COUNT/4) + 3, 
-                          5); 
+                          (RECORD_COUNT/4) + 3,
+                          5);
 
         /* Reopen, db should be populated. */
         openDb(null, DB_NAME1);
@@ -266,7 +333,6 @@ public class TruncateAndRemoveTest extends TestCase {
         closeEnv();
         batchCleanAndVerify(saveId);
     }
-
     /**
      * Test that remove generates the right number of obsolete LNs.
      */
@@ -276,14 +342,18 @@ public class TruncateAndRemoveTest extends TestCase {
         openEnv(true);
         openDb(null, DB_NAME1);
         writeAndCountRecords(null, RECORD_COUNT);
-        DatabaseId saveId = DbInternal.dbGetDatabaseImpl(db).getId();
-        db.close();
-        db = null;
+        DatabaseImpl saveDb = dbImpl;
+        DatabaseId saveId = dbImpl.getId();
+        closeDb();
 
         Transaction txn = env.beginTransaction(null, null);
         env.removeDatabase(txn, DB_NAME1);
         ObsoleteCounts beforeCommit = getObsoleteCounts();
         txn.commit();
+        truncateOrRemoveDone = true;
+
+        /* Make sure use count is decremented when we commit. */
+        assertDbInUse(saveDb, false);
 
         verifyUtilization(beforeCommit,
                           /* LNs + old NameLN, old MapLN, delete MapLN */
@@ -306,11 +376,15 @@ public class TruncateAndRemoveTest extends TestCase {
         openEnv(false);
         openDb(null, DB_NAME1);
         writeAndCountRecords(null, RECORD_COUNT);
-        DatabaseId saveId = DbInternal.dbGetDatabaseImpl(db).getId();
-        db.close();
-        db = null;
+        DatabaseImpl saveDb = dbImpl;
+        DatabaseId saveId = dbImpl.getId();
+        closeDb();
         ObsoleteCounts beforeOperation = getObsoleteCounts();
         env.removeDatabase(null, DB_NAME1);
+        truncateOrRemoveDone = true;
+
+        /* Make sure use count is decremented. */
+        assertDbInUse(saveDb, false);
 
         verifyUtilization(beforeOperation,
                           /* LNs + new NameLN, old NameLN, old MapLN, delete
@@ -335,13 +409,16 @@ public class TruncateAndRemoveTest extends TestCase {
         openEnv(true);
         openDb(null, DB_NAME1);
         writeAndCountRecords(null, RECORD_COUNT);
-        db.close();
-        db = null;
+        DatabaseImpl saveDb = dbImpl;
+        closeDb();
         Transaction txn = env.beginTransaction(null, null);
         env.removeDatabase(txn, DB_NAME1);
         ObsoleteCounts beforeAbort = getObsoleteCounts();
         txn.abort();
 
+        /* Make sure use count is decremented when we abort. */
+        assertDbInUse(saveDb, false);
+
         verifyUtilization(beforeAbort, 0, 0);
 
         /* All records should be there. */
@@ -350,7 +427,7 @@ public class TruncateAndRemoveTest extends TestCase {
 
         closeEnv();
 
-        /* 
+        /*
          * Batch clean and then check the record count again, just to make sure
          * we don't lose any valid data.
          */
@@ -368,6 +445,17 @@ public class TruncateAndRemoveTest extends TestCase {
         closeEnv();
     }
 
+    /**
+     * The same as testRemoveNotResident but forces fetching of obsolets LNs
+     * in order to count their sizes accurately.
+     */
+    public void testRemoveNotResidentFetchObsoleteSize()
+        throws Exception {
+
+        fetchObsoleteSize = true;
+        testRemoveNotResident();
+    }
+
     /**
      * Test that we can properly account for a non-resident database.
      */
@@ -379,12 +467,9 @@ public class TruncateAndRemoveTest extends TestCase {
         openDb(null, DB_NAME1);
         writeAndCountRecords(null, RECORD_COUNT);
         DatabaseId saveId = DbInternal.dbGetDatabaseImpl(db).getId();
-        db.close();
-        db = null;
-        env.close();
-        env = null;
+        closeEnv();
 
-        /* 
+        /*
          * Open the environment and remove the database. The
          * database is not resident at all.
          */
@@ -393,13 +478,14 @@ public class TruncateAndRemoveTest extends TestCase {
         env.removeDatabase(txn, DB_NAME1);
         ObsoleteCounts beforeCommit = getObsoleteCounts();
         txn.commit();
+        truncateOrRemoveDone = true;
 
         verifyUtilization(beforeCommit,
                           /* LNs + old NameLN, old MapLN, delete MapLN */
                           RECORD_COUNT + 3,
-                          /* 15 IN for data tree, 
-                             2 for file summary db, 2 for re-logged INs */
-                          15 + 2 + 2);
+                          /* 15 INs for data tree */
+                          15,
+                          true);
 
         /* check record count. */
         openDb(null, DB_NAME1);
@@ -409,6 +495,17 @@ public class TruncateAndRemoveTest extends TestCase {
         batchCleanAndVerify(saveId);
     }
 
+    /**
+     * The same as testRemovePartialResident but forces fetching of obsolets
+     * LNs in order to count their sizes accurately.
+     */
+    public void testRemovePartialResidentFetchObsoleteSize()
+        throws Exception {
+
+        fetchObsoleteSize = true;
+        testRemovePartialResident();
+    }
+
     /**
      * Test that we can properly account for partially resident tree.
      */
@@ -420,34 +517,36 @@ public class TruncateAndRemoveTest extends TestCase {
         openDb(null, DB_NAME1);
         writeAndCountRecords(null, RECORD_COUNT);
         DatabaseId saveId = DbInternal.dbGetDatabaseImpl(db).getId();
-        db.close();
-        db = null;
-        env.close();
-        env = null;
+        closeEnv();
 
-        /* 
+        /*
          * Open the environment and remove the database. Pull 1 BIN in.
          */
         openEnv(true);
         openDb(null, DB_NAME1);
         Cursor c = db.openCursor(null, null);
         assertEquals(OperationStatus.SUCCESS,
-                     c.getFirst(new DatabaseEntry(), new DatabaseEntry(), 
+                     c.getFirst(new DatabaseEntry(), new DatabaseEntry(),
                                 LockMode.DEFAULT));
         c.close();
-        db.close();
-        db = null;
+        DatabaseImpl saveDb = dbImpl;
+        closeDb();
 
         Transaction txn = env.beginTransaction(null, null);
         env.removeDatabase(txn, DB_NAME1);
         ObsoleteCounts beforeCommit = getObsoleteCounts();
         txn.commit();
+        truncateOrRemoveDone = true;
+
+        /* Make sure use count is decremented when we commit. */
+        assertDbInUse(saveDb, false);
 
         verifyUtilization(beforeCommit,
                           /* LNs + old NameLN, old MapLN, delete MapLN */
                           RECORD_COUNT + 3,
-                          /* 15 IN for data tree, 2 for file summary db */
-                          15 + 2);
+                          /* 15 INs for data tree */
+                          15,
+                          true);
 
         /* check record count. */
         openDb(null, DB_NAME1);
@@ -456,7 +555,7 @@ public class TruncateAndRemoveTest extends TestCase {
         closeEnv();
         batchCleanAndVerify(saveId);
     }
-  
+
     /**
      * Tests that a log file is not deleted by the cleaner when it contains
      * entries in a database that is pending deletion.
@@ -464,22 +563,22 @@ public class TruncateAndRemoveTest extends TestCase {
     public void testDBPendingDeletion()
         throws DatabaseException, InterruptedException {
 
-        doDBPendingTest(RECORD_COUNT, false /*deleteAll*/, 7);
+        doDBPendingTest(RECORD_COUNT + 30, false /*deleteAll*/, 5);
     }
 
     /**
      * Like testDBPendingDeletion but creates a scenario where only a single
      * log file is cleaned, and that log file contains only known obsolete
-     * log entries.  This reproduced a bug where we neglected to adding
-     * pending deleted DBs to the cleaner's pending DB set if all entries in
-     * the log file were known obsoleted. [#13333]
+     * log entries.  This reproduced a bug where we neglected to add pending
+     * deleted DBs to the cleaner's pending DB set if all entries in the log
+     * file were known obsoleted. [#13333]
      */
     public void testObsoleteLogFile()
         throws DatabaseException, InterruptedException {
 
-        doDBPendingTest(40, true /*deleteAll*/, 1);
+        doDBPendingTest(70, true /*deleteAll*/, 1);
     }
-    
+
     private void doDBPendingTest(long recordCount,
                                  boolean deleteAll,
                                  int expectFilesCleaned)
@@ -493,14 +592,14 @@ public class TruncateAndRemoveTest extends TestCase {
         long remainingRecordCount = deleteAll ? 0 : recordCount;
         env.checkpoint(FORCE_CHECKPOINT);
         ObsoleteCounts obsoleteCounts = getObsoleteCounts();
-        DatabaseImpl dbImpl = DbInternal.dbGetDatabaseImpl(db); 
-        db.close();
-        db = null;
-        assertTrue(!dbImpl.isDeleteFinished());
-        assertTrue(!dbImpl.isDeleted());
+        DatabaseImpl saveDb = dbImpl;
+        closeDb();
+        assertTrue(!saveDb.isDeleteFinished());
+        assertTrue(!saveDb.isDeleted());
+        assertDbInUse(saveDb, false);
 
         /* Make sure that we wrote a full file's worth of LNs. */
-        assertTrue(logFiles.size() >= 3);
+        assertTrue(logFiles.size() >= 2);
         assertTrue(logFilesExist(logFiles));
 
         /* Remove the database but do not commit yet. */
@@ -509,9 +608,10 @@ public class TruncateAndRemoveTest extends TestCase {
 
         /* The obsolete count should be <= 1 (for the NameLN). */
         obsoleteCounts = verifyUtilization(obsoleteCounts, 1, 0);
+        truncateOrRemoveDone = true;
 
         junitThread = new JUnitThread("Committer") {
-            public void testBody() 
+            public void testBody()
                 throws DatabaseException {
                 try {
                     txn.commit();
@@ -528,11 +628,7 @@ public class TruncateAndRemoveTest extends TestCase {
          */
         final Object lock = new Object();
 
-        dbImpl.setPendingDeletedHook(new TestHook() {
-            public void doIOHook()
-                throws IOException {
-                throw new UnsupportedOperationException();
-            }
+        saveDb.setPendingDeletedHook(new TestHook() {
             public void doHook() {
                 synchronized (lock) {
                     try {
@@ -545,7 +641,13 @@ public class TruncateAndRemoveTest extends TestCase {
                 }
             }
             public Object getHookValue() {
-        	return null;
+                throw new UnsupportedOperationException();
+            }
+            public void doIOHook() throws IOException {
+                throw new UnsupportedOperationException();
+            }
+            public void hookSetup() {
+                throw new UnsupportedOperationException();
             }
         });
 
@@ -554,8 +656,9 @@ public class TruncateAndRemoveTest extends TestCase {
             junitThread.start();
             lock.wait();
         }
-        assertTrue(!dbImpl.isDeleteFinished());
-        assertTrue(dbImpl.isDeleted());
+        assertTrue(!saveDb.isDeleteFinished());
+        assertTrue(saveDb.isDeleted());
+        assertDbInUse(saveDb, true);
 
         /* Expect obsolete LNs: NameLN */
         obsoleteCounts = verifyUtilization(obsoleteCounts, 1, 0);
@@ -582,8 +685,9 @@ public class TruncateAndRemoveTest extends TestCase {
             e.printStackTrace();
             fail(e.toString());
         }
-        assertTrue(dbImpl.isDeleteFinished());
-        assertTrue(dbImpl.isDeleted());
+        assertTrue(saveDb.isDeleteFinished());
+        assertTrue(saveDb.isDeleted());
+        assertDbInUse(saveDb, false);
 
         /* Expect obsolete LNs: recordCount + MapLN + FSLNs (apprx). */
         verifyUtilization(obsoleteCounts, remainingRecordCount + 6, 0);
@@ -594,6 +698,155 @@ public class TruncateAndRemoveTest extends TestCase {
         assertTrue(!logFilesExist(logFiles));
     }
 
+    /*
+     * The xxxForceTreeWalk tests set the DatabaseImpl
+     * forceTreeWalkForTruncateAndRemove field to true, which will force a walk
+     * of the tree to count utilization during truncate/remove, rather than
+     * using the per-database info.  This is used to test the "old technique"
+     * for counting utilization, which is now used only if the database was
+     * created prior to log version 6.
+     */
+
+    public void testTruncateForceTreeWalk()
+        throws Exception {
+
+        DatabaseImpl.forceTreeWalkForTruncateAndRemove = true;
+        try {
+            testTruncate();
+        } finally {
+            DatabaseImpl.forceTreeWalkForTruncateAndRemove = false;
+        }
+    }
+
+    public void testTruncateAbortForceTreeWalk()
+        throws Exception {
+
+        DatabaseImpl.forceTreeWalkForTruncateAndRemove = true;
+        try {
+            testTruncateAbort();
+        } finally {
+            DatabaseImpl.forceTreeWalkForTruncateAndRemove = false;
+        }
+    }
+
+    public void testTruncateRepopulateAbortForceTreeWalk()
+        throws Exception {
+
+        DatabaseImpl.forceTreeWalkForTruncateAndRemove = true;
+        try {
+            testTruncateRepopulateAbort();
+        } finally {
+            DatabaseImpl.forceTreeWalkForTruncateAndRemove = false;
+        }
+    }
+
+    public void testRemoveForceTreeWalk()
+        throws Exception {
+
+        DatabaseImpl.forceTreeWalkForTruncateAndRemove = true;
+        try {
+            testRemove();
+        } finally {
+            DatabaseImpl.forceTreeWalkForTruncateAndRemove = false;
+        }
+    }
+
+
+    public void testNonTxnalRemoveForceTreeWalk()
+        throws Exception {
+
+        DatabaseImpl.forceTreeWalkForTruncateAndRemove = true;
+        try {
+            testNonTxnalRemove();
+        } finally {
+            DatabaseImpl.forceTreeWalkForTruncateAndRemove = false;
+        }
+    }
+
+    public void testRemoveAbortForceTreeWalk()
+        throws Exception {
+
+        DatabaseImpl.forceTreeWalkForTruncateAndRemove = true;
+        try {
+            testRemoveAbort();
+        } finally {
+            DatabaseImpl.forceTreeWalkForTruncateAndRemove = false;
+        }
+    }
+
+    public void testRemoveNotResidentForceTreeWalk()
+        throws Exception {
+
+        DatabaseImpl.forceTreeWalkForTruncateAndRemove = true;
+        try {
+            testRemoveNotResident();
+        } finally {
+            DatabaseImpl.forceTreeWalkForTruncateAndRemove = false;
+        }
+    }
+
+    public void testRemovePartialResidentForceTreeWalk()
+        throws Exception {
+
+        DatabaseImpl.forceTreeWalkForTruncateAndRemove = true;
+        try {
+            testRemovePartialResident();
+        } finally {
+            DatabaseImpl.forceTreeWalkForTruncateAndRemove = false;
+        }
+    }
+
+    public void testDBPendingDeletionForceTreeWalk()
+        throws Exception {
+
+        DatabaseImpl.forceTreeWalkForTruncateAndRemove = true;
+        try {
+            testDBPendingDeletion();
+        } finally {
+            DatabaseImpl.forceTreeWalkForTruncateAndRemove = false;
+        }
+    }
+
+    public void testObsoleteLogFileForceTreeWalk()
+        throws Exception {
+
+        DatabaseImpl.forceTreeWalkForTruncateAndRemove = true;
+        try {
+            testObsoleteLogFile();
+        } finally {
+            DatabaseImpl.forceTreeWalkForTruncateAndRemove = false;
+        }
+    }
+
+    /**
+     * Tickles a bug that caused NPE during recovery during the sequence:
+     * delete record, trucate DB, crash (close without checkpoint), and
+     * recover. [#16515]
+     */
+    public void testDeleteTruncateRecover()
+        throws DatabaseException {
+
+        /* Delete a record. */
+        openEnv(true);
+        openDb(null, DB_NAME1);
+        writeAndCountRecords(null, 1);
+        closeDb();
+
+        /* Truncate DB. */
+        Transaction txn = env.beginTransaction(null, null);
+        truncate(txn, false);
+        txn.commit();
+
+        /* Close without checkpoint. */
+        envImpl.close(false /*doCheckpoint*/);
+        envImpl = null;
+        env = null;
+
+        /* Recover -- the bug cause NPE here. */
+        openEnv(true);
+        closeEnv();
+    }
+
     private void writeAndCountRecords(Transaction txn, long count)
         throws DatabaseException {
 
@@ -669,8 +922,7 @@ public class TruncateAndRemoveTest extends TestCase {
     }
 
     /* Truncate database and check the count. */
-    private void truncate(Transaction useTxn,
-                          boolean getCount)
+    private void truncate(Transaction useTxn, boolean getCount)
         throws DatabaseException {
 
         long nTruncated = env.truncateDatabase(useTxn, DB_NAME1, getCount);
@@ -681,7 +933,7 @@ public class TruncateAndRemoveTest extends TestCase {
 
         assertEquals(0, countRecords(useTxn));
     }
-                          
+
     /**
      * Returns how many records are in the database.
      */
@@ -690,21 +942,26 @@ public class TruncateAndRemoveTest extends TestCase {
 
         DatabaseEntry key = new DatabaseEntry();
         DatabaseEntry data = new DatabaseEntry();
+        boolean opened = false;
         if (db == null) {
             openDb(useTxn, DB_NAME1);
+            opened = true;
         }
         Cursor cursor = db.openCursor(useTxn, null);
+        int count = 0;
         try {
-            int count = 0;
             OperationStatus status = cursor.getFirst(key, data, null);
             while (status == OperationStatus.SUCCESS) {
                 count += 1;
                 status = cursor.getNext(key, data, null);
             }
-            return count;
         } finally {
             cursor.close();
         }
+        if (opened) {
+            closeDb();
+        }
+        return count;
     }
 
     /**
@@ -715,42 +972,73 @@ public class TruncateAndRemoveTest extends TestCase {
         throws DatabaseException {
 
         FileSummary[] files = (FileSummary[])
-            DbInternal.envGetEnvironmentImpl(env)
-                      .getUtilizationProfile()
-                      .getFileSummaryMap(true)
-                      .values().toArray(new FileSummary[0]);
+            envImpl.getUtilizationProfile()
+                   .getFileSummaryMap(true)
+                   .values().toArray(new FileSummary[0]);
         int lnCount = 0;
         int inCount = 0;
+        int lnSize = 0;
+        int lnSizeCounted = 0;
         for (int i = 0; i < files.length; i += 1) {
             lnCount += files[i].obsoleteLNCount;
             inCount += files[i].obsoleteINCount;
+            lnSize += files[i].obsoleteLNSize;
+            lnSizeCounted += files[i].obsoleteLNSizeCounted;
         }
 
-        return new ObsoleteCounts(lnCount, inCount);
+        return new ObsoleteCounts(lnCount, inCount, lnSize, lnSizeCounted);
     }
 
     private class ObsoleteCounts {
         int obsoleteLNs;
         int obsoleteINs;
+        int obsoleteLNSize;
+        int obsoleteLNSizeCounted;
 
-        ObsoleteCounts(int obsoleteLNs, int obsoleteINs) {
+        ObsoleteCounts(int obsoleteLNs,
+                       int obsoleteINs,
+                       int obsoleteLNSize,
+                       int obsoleteLNSizeCounted) {
             this.obsoleteLNs = obsoleteLNs;
             this.obsoleteINs = obsoleteINs;
+            this.obsoleteLNSize = obsoleteLNSize;
+            this.obsoleteLNSizeCounted = obsoleteLNSizeCounted;
         }
 
         public String toString() {
-            return "lns=" + obsoleteLNs + " ins=" + obsoleteINs;
+            return "lns=" + obsoleteLNs + " ins=" + obsoleteINs +
+                   " lnSize=" + obsoleteLNSize +
+                   " lnSizeCounted=" + obsoleteLNSizeCounted;
         }
     }
 
-    /* 
-     * Check obsolete counts. If the expected IN count is negative, don't
-     * check.
+    private ObsoleteCounts verifyUtilization(ObsoleteCounts prev,
+                                             long expectedLNs,
+                                             int expectedINs)
+        throws DatabaseException {
+
+        return verifyUtilization(prev, expectedLNs, expectedINs, false);
+    }
+
+    /*
+     * Check obsolete counts. If the expected IN count is zero, don't
+     * check the obsolete IN count.  Always check the obsolete LN count.
      */
     private ObsoleteCounts verifyUtilization(ObsoleteCounts prev,
                                              long expectedLNs,
-                                             int expectedINs) 
+                                             int expectedINs,
+                                             boolean expectNonResident)
         throws DatabaseException {
+
+        /*
+         * If we are not forcing a tree walk OR all nodes are resident OR we
+         * have explicitly configured fetchObsoleteSize, then the size of every
+         * LN should have been counted.
+         */
+        boolean expectAccurateObsoleteLNSize =
+            !DatabaseImpl.forceTreeWalkForTruncateAndRemove ||
+            !expectNonResident ||
+            fetchObsoleteSize;
 	
         ObsoleteCounts now = getObsoleteCounts();
         String beforeAndAfter = "before: " + prev + " now: " + now;
@@ -759,15 +1047,48 @@ public class TruncateAndRemoveTest extends TestCase {
         }
 
         assertEquals(beforeAndAfter, expectedLNs,
-                     now.obsoleteLNs - prev.obsoleteLNs);
+		     now.obsoleteLNs - prev.obsoleteLNs);
+        if (expectedLNs > 0) {
+            int size = now.obsoleteLNSize - prev.obsoleteLNSize;
+            int counted = now.obsoleteLNSizeCounted -
+                          prev.obsoleteLNSizeCounted;
+            assertTrue(String.valueOf(size), size > 0);
+
+            if (expectAccurateObsoleteLNSize) {
+                assertEquals(beforeAndAfter, counted,
+			     now.obsoleteLNs - prev.obsoleteLNs);
+            }
+        }
         if (expectedINs > 0) {
             assertEquals(beforeAndAfter, expectedINs,
                          now.obsoleteINs - prev.obsoleteINs);
         }
 
+        /*
+         * We pass expectAccurateDbUtilization as false when
+         * truncateOrRemoveDone, because the database utilization info for that
+         * database is now gone.
+         */
+        VerifyUtils.verifyUtilization
+            (envImpl,
+             true,                   // expectAccurateObsoleteLNCount
+             expectAccurateObsoleteLNSize,
+             !truncateOrRemoveDone); // expectAccurateDbUtilization
+
         return now;
     }
 
+    /**
+     * Checks whether a given DB has a non-zero use count.  Does nothing if
+     * je.dbEviction is not enabled, since reference counts are only maintained
+     * if that config parameter is enabled.
+     */
+    private void assertDbInUse(DatabaseImpl db, boolean inUse) {
+        if (dbEviction) {
+            assertEquals(inUse, db.isInUse());
+        }
+    }
+
     /**
      * Returns true if all files exist, or false if any file is deleted.
      */
@@ -789,18 +1110,18 @@ public class TruncateAndRemoveTest extends TestCase {
     /*
      * Run batch cleaning and verify that there are no files with these
      * log entries.
-     */ 
-    private void batchCleanAndVerify(DatabaseId dbId) 
+     */
+    private void batchCleanAndVerify(DatabaseId dbId)
         throws Exception {
 
-        /* 
+        /*
          * Open the environment, flip the log files to reduce mixing of new
          * records and old records and add more records to force the
          * utilization level of the removed records down.
          */
         openEnv(true);
         openDb(null, DB_NAME2);
-        long lsn = DbInternal.envGetEnvironmentImpl(env).forceLogFileFlip();
+        long lsn = envImpl.forceLogFileFlip();
         CheckpointConfig force = new CheckpointConfig();
         force.setForce(true);
         env.checkpoint(force);
@@ -808,11 +1129,10 @@ public class TruncateAndRemoveTest extends TestCase {
         writeAndCountRecords(null, RECORD_COUNT * 3);
         env.checkpoint(force);
 
-        db.close();
-        db = null;
+        closeDb();
 
         /* Check log files, there should be entries with this database. */
-        CheckReader checker = new CheckReader(env, dbId, true);
+        CheckReader checker = new CheckReader(envImpl, dbId, true);
         while (checker.readNextEntry()) {
         }
 
@@ -835,12 +1155,11 @@ public class TruncateAndRemoveTest extends TestCase {
         }
 
         /* Check log files, there should be no entries with this database. */
-        checker = new CheckReader(env, dbId, false);
+        checker = new CheckReader(envImpl, dbId, false);
         while (checker.readNextEntry()) {
         }
 
         closeEnv();
-        
     }
 
     class CheckReader extends DumpFileReader{
@@ -855,31 +1174,30 @@ public class TruncateAndRemoveTest extends TestCase {
          * with this database id. If true, the log should have entries
          * with this database id.
          */
-        CheckReader(Environment env, DatabaseId dbId, boolean expectEntries)
+        CheckReader(EnvironmentImpl envImpl,
+                    DatabaseId dbId,
+                    boolean expectEntries)
             throws DatabaseException, IOException {
-            
-            super(DbInternal.envGetEnvironmentImpl(env),
-        	  1000, DbLsn.NULL_LSN, DbLsn.NULL_LSN,
+
+            super(envImpl, 1000, DbLsn.NULL_LSN, DbLsn.NULL_LSN,
                   null, null, false);
             this.dbId = dbId;
             this.expectEntries = expectEntries;
         }
-        
-        protected boolean processEntry(ByteBuffer entryBuffer) 
-            throws DatabaseException {
-            
-            /* Figure out what kind of log entry this is */
-            LogEntryType lastEntryType =
-                LogEntryType.findType(currentEntryTypeNum,
-                                      currentEntryTypeVersion);
-            boolean isNode = LogEntryType.isNodeType(currentEntryTypeNum,
-                                                     currentEntryTypeVersion);
 
+        protected boolean processEntry(ByteBuffer entryBuffer)
+            throws DatabaseException {
+
+            /* Figure out what kind of log entry this is */
+            byte type = currentEntryHeader.getType();
+            LogEntryType lastEntryType = LogEntryType.findType(type);
+            boolean isNode = LogEntryType.isNodeType(type);
 
             /* Read the entry. */
             LogEntry entry = lastEntryType.getSharedLogEntry();
-            entry.readEntry(entryBuffer, currentEntrySize,
-                            currentEntryTypeVersion, true);
+            entry.readEntry(currentEntryHeader,
+            		        entryBuffer,
+            		        true); // readFullItem
 
             long lsn = getLastLsn();
             if (isNode) {
